/* vim: set expandtab ts=4 sw=4: */
/*
 * You may redistribute this program and/or modify it under the terms of
 * the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
var Fs = require('fs');
var nThen = require('nthen');
var Codestyle = require('./Codestyle');
var Cp = require('./Cp');
var Spawn = require('child_process').spawn;
var Os = require('os');

var WORKERS = Math.floor(Os.cpus().length * 1.25);

process.on('exit', function () {
    console.log("Total build time: " + Math.floor(process.uptime() * 1000) + "ms.");
});

var Builder = require('./builder');

Builder.configure({
    rebuildIfChanges: Fs.readFileSync(__filename).toString('utf8'),
    buildDir: 'buildjs'
}, function(builder, waitFor) {

    // ['linux','darwin','sunos','win32','freebsd']
    var system = builder.config.systemName = 'linux';
    builder.config.gcc = 'gcc';

    builder.config.tempDir = '/tmp';
    builder.config.useTempFiles = true;
    builder.config.cflags.push(
        '-std=c99',
        '-Wall',
        '-Wextra',
        '-Werror',
        '-Wno-pointer-sign',
        '-pedantic',
        '-D',builder.config.systemName + '=1',
        '-DHAS_ETH_INTERFACE=1',
        '-Wno-unused-parameter',
        '-Wno-unused-result',

        // Broken GCC patch makes -fstack-protector-all not work
        // workaround is to give -fno-stack-protector first.
        // see: https://bugs.launchpad.net/ubuntu/+source/gcc-4.5/+bug/691722
        '-fno-stack-protector',
        '-fstack-protector-all',
        '-Wstack-protector',

        '-D','HAS_BUILTIN_CONSTANT_P',

        '-g',
        '-fPIE',
        '-pie',

//        '-flto', not available on some  machines

        '-D','CJDNS_MAX_PEERS=256',

        // disable for speed, enable for safety
        '-D','Log_DEBUG',
        '-D','Identity_CHECK=1',
        '-D','Allocator_USE_CANARIES=1',
        '-D','PARANOIA=1'
    );
    if (system === 'win32') {
        builder.config.cflags.push(
            '!-fPIE',
            '!-pie',
            '-Wno-format',
            '!-DHAS_ETH_INTERFACE=1'
        );
        builder.config.libs.push(
            '-lssp'
        );
    } else {
        builder.config.ldflags.push(
            '-Wl,-z,relro,-z,now,-z,noexecstack',
            '-lrt'
        );
    }

    // Build dependencies
    nThen(function (waitFor) {
        Fs.exists('buildjs/dependencies', waitFor(function (exists) {
            if (exists) { return; }
            console.log("Copy dependencies");
            Cp('./node_build/dependencies', './buildjs/dependencies', waitFor());
        }));
    }).nThen(function (waitFor) {
        builder.config.libs.push(
            'buildjs/dependencies/cnacl/jsbuild/libnacl.a'
        );
        builder.config.includeDirs.push(
            'buildjs/dependencies/cnacl/jsbuild/include/'
        );
        Fs.exists('buildjs/dependencies/cnacl/jsbuild/libnacl.a', waitFor(function (exists) {
            if (exists) { return; }
            console.log("Build NaCl");
            var cwd = process.cwd();
            process.chdir('./buildjs/dependencies/cnacl/');
            var NaCl = require(process.cwd() + '/node_build/make.js');
            NaCl.build(function (args, callback) {
                if (builder.config.systemName !== 'win32') { args.unshift('-fPIC'); }
                args.unshift('-O2', '-fomit-frame-pointer');
                builder.cc(args, callback);
            }, waitFor(function () {
                process.chdir(cwd);
            }));
        }));

    }).nThen(function (waitFor) {
        builder.config.libs.push(
            'buildjs/dependencies/libuv/libuv.a',
            '-lpthread'
        );
        if (builder.config.systemName === 'win32') {
            builder.config.libs.push(
                '-lws2_32',
                '-lpsapi',   // GetProcessMemoryInfo()
                '-liphlpapi' // GetAdapterAddresses()
            );
        }
        builder.config.includeDirs.push(
            'buildjs/dependencies/libuv/include/'
        );
        Fs.exists('buildjs/dependencies/libuv/libuv.a', waitFor(function (exists) {
            if (exists) { return; }
            console.log("Build Libuv");
            var cwd = process.cwd();
            process.chdir('./buildjs/dependencies/libuv/');
            var args = ['-j', WORKERS, 'CFLAGS=-fPIC', 'CC='+builder.config.gcc]
            if (builder.config.systemName === 'win32') { args.push('PLATFORM=mingw32'); }
            var make = Spawn('make', args);
            make.stdout.on('data', function(dat) { process.stdout.write(dat.toString()); });
            make.stderr.on('data', function(dat) { process.stderr.write(dat.toString()); });
            make.on('close', waitFor(function () {
                process.chdir(cwd);
            }));
        }));

    }).nThen(waitFor());

}).build(function (builder, waitFor) {

    builder.buildExecutable('admin/angel/cjdroute2.c', './buildjs/cjdroute', waitFor());
    builder.buildExecutable('test/testcjdroute.c', './buildjs/testcjdroute', waitFor());

    if (process.argv.indexOf('--contrib') > -1) {
        builder.buildExecutable('contrib/c/publictoip6.c', './publictoip6', waitFor());
        builder.buildExecutable('contrib/c/privatetopublic.c', './privatetopublic', waitFor());
        builder.buildExecutable('contrib/c/sybilsim.c', './sybilsim', waitFor());
        builder.buildExecutable('contrib/c/benc2json.c', './benc2json', waitFor());
        builder.buildExecutable('contrib/c/cleanconfig.c', './cleanconfig', waitFor());
        builder.buildExecutable('contrib/c/dnsserv.c', './dnsserv', waitFor());
    }

}).test(function (builder, waitFor) {

    nThen(function (waitFor) {

        var out = '';
        var test = Spawn('./buildjs/testcjdroute');
        test.stdout.on('data', function(dat) { out += dat.toString(); });
        test.stderr.on('data', function(dat) { process.stderr.write(dat.toString()); });
        test.on('close', waitFor(function (ret) {
            if (ret !== 0) {
                console.log(out);
                console.log('\033[1;31mFailed to build cjdns.\033[0m');
                waitFor.abort();
            } else {
                console.log('\033[1;32mBuild completed successfully, type ./cjdroute to begin setup.\033[0m');
            }
        }));

    }).nThen(function (waitFor) {

        console.log("Checking codestyle");
        var output = '';
        nThen(function (waitFor) {
            builder.rebuiltFiles.forEach(function (fileName) {
                console.log("Checking " + fileName);
                if (fileName.indexOf('/dependencies/') !== -1) { return; }
                Codestyle.checkFile(fileName, waitFor(function (out) {
                    output += out;
                }));
            });

        }).nThen(function (waitFor) {
            if (output !== '') {
                throw new Error("Codestyle failure\n" + output);
            }

        }).nThen(waitFor());

    }).nThen(waitFor());

}).pack(function (builder, waitFor) {

    Fs.exists('./buildjs/cjdroute', waitFor(function (exists) {
        if (!exists) { return; }
        Fs.rename('./buildjs/cjdroute', './cjdroute', waitFor(function (err) {
            if (err) { throw err; }
        }));
    }));

});
